import hashlib
import itertools
import math
import os
import random
import struct
import sys
from functools import lru_cache

from sneks.engine.config.definition import Colors
from sneks.engine.core.cell import Cell
from sneks.engine.engine import cells
from sneks.engine.engine.cells import get_relative_to
from sneks.engine.engine.mover import Asteroid

try:
    import pygame
    from pygame import Surface
except ModuleNotFoundError:
    pygame = object  # type: ignore
    Surface = object  # type: ignore

from sneks.engine.config.instantiation import config
from sneks.engine.gui.recorder import Recorder

assert config.graphics is not None

ROWS = config.game.rows
COLUMNS = config.game.columns
CELL_SIZE = config.graphics.cell_size
PADDING = config.graphics.padding
COLOR_BORDER = config.graphics.colors.border
COLOR_BACKGROUND = config.graphics.colors.background
COLOR_INVALID = config.graphics.colors.invalid

HEIGHT = (2 + ROWS) * CELL_SIZE + ROWS * PADDING
WIDTH = (2 + COLUMNS) * CELL_SIZE + COLUMNS * PADDING


class Painter:
    screen: Surface | None = None

    def __init__(self, recorder: Recorder | None = None):
        self.recorder = recorder

    def initialize(self):
        if config.graphics.headless:
            os.environ["SDL_VIDEODRIVER"] = "dummy"
        pygame.init()
        self.screen = pygame.display.set_mode((WIDTH, HEIGHT))
        pygame.display.set_caption("Sneks on a Toroidal Plane")

    @staticmethod
    def get_rect(surface: pygame.Surface, cell: Cell) -> pygame.Rect:
        return surface.get_rect(
            top=CELL_SIZE
            + PADDING
            + (config.game.rows - cell.y - 1) * (CELL_SIZE + PADDING),
            left=CELL_SIZE + PADDING + cell.x * (CELL_SIZE + PADDING),
        )

    def draw_snake(self, head: Cell, cells, alive: bool, color: tuple[int, int, int]):
        assert self.screen is not None
        surface = pygame.Surface((CELL_SIZE - PADDING, CELL_SIZE - PADDING))
        fill_horizontal = pygame.Surface((PADDING * 2, CELL_SIZE - PADDING))
        fill_vertical = pygame.Surface((CELL_SIZE - PADDING, PADDING * 2))
        surface.fill(color)
        fill_horizontal.fill(color)
        fill_vertical.fill(color)
        previous: Cell | None = None
        previous_color: tuple[int, ...] = color
        for cell in reversed(cells):
            surface.fill(previous_color)
            fill_horizontal.fill(previous_color)
            fill_vertical.fill(previous_color)
            rect = self.get_rect(surface=surface, cell=cell)
            self.screen.blit(surface, rect)
            if previous is not None:
                # This is super verbose and can be refactored
                looped = abs(cell.x - previous.x) + abs(cell.y - previous.y) > 1

                relative = get_relative_to(cell, previous)
                dx = min(1, max(relative.x, -1))
                dy = min(1, max(relative.y, -1))
                if looped:
                    dx *= -1
                    dy *= -1
                match (dx, dy):
                    case (1, 0):
                        # fill the padding to the left
                        top = (
                            CELL_SIZE
                            + (config.game.rows - cell.y - 1) * (CELL_SIZE + PADDING)
                            + PADDING
                        )
                        left = CELL_SIZE + cell.x * (CELL_SIZE + PADDING) - PADDING
                        rect = fill_horizontal.get_rect(top=top, left=left)
                        self.screen.blit(fill_horizontal, rect)
                        if looped:
                            # fill the padding to the right of the previous
                            top = (
                                CELL_SIZE
                                + (config.game.rows - cell.y - 1)
                                * (CELL_SIZE + PADDING)
                                + PADDING
                            )
                            left = (
                                CELL_SIZE
                                + (previous.x + 1) * (CELL_SIZE + PADDING)
                                - PADDING
                            )
                            rect = fill_horizontal.get_rect(top=top, left=left)
                            self.screen.blit(fill_horizontal, rect)
                    case (-1, 0):
                        # fill the padding to the right
                        top = (
                            CELL_SIZE
                            + (config.game.rows - cell.y - 1) * (CELL_SIZE + PADDING)
                            + PADDING
                        )
                        left = CELL_SIZE + previous.x * (CELL_SIZE + PADDING) - PADDING
                        rect = fill_horizontal.get_rect(top=top, left=left)
                        self.screen.blit(fill_horizontal, rect)
                        if looped:
                            # fill the padding to the right of the current
                            top = (
                                CELL_SIZE
                                + (config.game.rows - cell.y - 1)
                                * (CELL_SIZE + PADDING)
                                + PADDING
                            )
                            left = (
                                CELL_SIZE
                                + (cell.x + 1) * (CELL_SIZE + PADDING)
                                - PADDING
                            )
                            rect = fill_horizontal.get_rect(top=top, left=left)
                            self.screen.blit(fill_horizontal, rect)
                    case (0, 1):
                        # fill the padding below
                        top = (
                            CELL_SIZE
                            + (config.game.rows - previous.y - 1)
                            * (CELL_SIZE + PADDING)
                            - PADDING
                        )
                        left = CELL_SIZE + cell.x * (CELL_SIZE + PADDING) + PADDING
                        rect = fill_vertical.get_rect(top=top, left=left)
                        self.screen.blit(fill_vertical, rect)
                        if looped:
                            # fill the padding to the below the current
                            top = (
                                CELL_SIZE
                                + (config.game.rows - cell.y) * (CELL_SIZE + PADDING)
                                - PADDING
                            )
                            left = CELL_SIZE + cell.x * (CELL_SIZE + PADDING) + PADDING
                            rect = fill_vertical.get_rect(top=top, left=left)
                            self.screen.blit(fill_vertical, rect)
                    case (0, -1):
                        # fill the padding above
                        top = (
                            CELL_SIZE
                            + (config.game.rows - cell.y - 1) * (CELL_SIZE + PADDING)
                            - PADDING
                        )
                        left = CELL_SIZE + cell.x * (CELL_SIZE + PADDING) + PADDING
                        rect = fill_vertical.get_rect(top=top, left=left)
                        self.screen.blit(fill_vertical, rect)
                        if looped:
                            # fill the padding to the below the previous
                            top = (
                                CELL_SIZE
                                + (config.game.rows - previous.y)
                                * (CELL_SIZE + PADDING)
                                - PADDING
                            )
                            left = CELL_SIZE + cell.x * (CELL_SIZE + PADDING) + PADDING
                            rect = fill_vertical.get_rect(top=top, left=left)
                            self.screen.blit(fill_vertical, rect)
            previous = cell
            previous_color = tuple(int(c * 0.9) for c in previous_color)
        if alive:
            surface.fill(
                struct.unpack(
                    "BBB", hashlib.md5(struct.pack("BBB", *color)).digest()[-3:]
                )
            )
            rect = self.get_rect(surface=surface, cell=head)
            self.screen.blit(surface, rect)

    def draw_asteroid(self, asteroid: Asteroid) -> None:
        assert self.screen is not None
        surface = pygame.Surface((CELL_SIZE + PADDING, CELL_SIZE + PADDING))
        # To show the asteroid tails

        # surface.fill(Colors.invalid)
        # for cell in asteroid.tail:
        #     rect = surface.get_rect(
        #         top=CELL_SIZE + (config.game.rows - cell.y - 1) * (CELL_SIZE + PADDING),
        #         left=CELL_SIZE + cell.x * (CELL_SIZE + PADDING),
        #     )
        #     self.screen.blit(surface, rect)
        surface.fill(Colors.asteroid)
        for cell in asteroid.cells:
            rect = surface.get_rect(
                top=CELL_SIZE + (config.game.rows - cell.y - 1) * (CELL_SIZE + PADDING),
                left=CELL_SIZE + cell.x * (CELL_SIZE + PADDING),
            )
            self.screen.blit(surface, rect)

    def draw_ended_head(self, head: Cell, splash: int):
        assert self.screen is not None
        surface = pygame.Surface((CELL_SIZE - PADDING, CELL_SIZE - PADDING))

        surface.fill(COLOR_INVALID)
        rect = self.get_rect(surface=surface, cell=head)
        self.screen.blit(surface, rect)

        surface.fill(tuple(int((0.85**splash) * c) for c in COLOR_INVALID))

        splash_cells = [
            cells.get_absolute_by_offset(cell=head, x_offset=x, y_offset=y)
            for x, y in itertools.product(
                range(-int(splash / 2), int(splash / 2)), repeat=2
            )
        ]
        for c in splash_cells:
            rect = self.get_rect(surface=surface, cell=c)
            self.screen.blit(surface, rect)

    def clear(self):
        self.screen.fill(COLOR_BACKGROUND)

    def draw_boarders(self):
        top = (0, 0, WIDTH, CELL_SIZE - PADDING)
        bottom = (0, HEIGHT - CELL_SIZE + PADDING, WIDTH, HEIGHT)
        left = (0, CELL_SIZE - PADDING, CELL_SIZE - PADDING, HEIGHT)
        right = (WIDTH - CELL_SIZE + PADDING, 0, WIDTH, HEIGHT)

        for rect in (top, bottom, left, right):
            pygame.draw.rect(self.screen, COLOR_BORDER, rect)

    @lru_cache
    def get_stars(self) -> list[tuple[int, int]]:
        return [
            (
                random.randint(CELL_SIZE + PADDING, WIDTH - CELL_SIZE - PADDING),
                random.randint(CELL_SIZE + PADDING, HEIGHT - CELL_SIZE - PADDING),
            )
            for _ in range(ROWS + COLUMNS)
        ]

    def draw_stars(self):
        for center in self.get_stars():
            modifier = math.prod(center) % (256 - max(COLOR_BACKGROUND))
            pygame.draw.circle(
                surface=self.screen,
                color=tuple(modifier + c for c in COLOR_BACKGROUND),
                center=center,
                radius=1.0,
            )

    def draw(self):
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                sys.exit()
        pygame.display.flip()
        if self.recorder:
            self.recorder.record_frame(self.screen)
        self.step_delay()

    def step_delay(self):
        if not config.graphics.headless:
            if config.graphics.step_keypress_wait:
                self.wait_for_keypress()
            pygame.time.delay(config.graphics.step_delay)

    def end_delay(self):
        if not config.graphics.headless:
            if config.graphics.end_keypress_wait:
                self.wait_for_keypress()
            pygame.time.delay(config.graphics.end_delay)

    @staticmethod
    def wait_for_keypress():
        while True:
            # allow the key to be held instead of waiting for each step
            if any(pygame.key.get_pressed()):
                break
            event = pygame.event.wait()
            if event.type == pygame.QUIT:
                pygame.quit()
                sys.exit()
            elif event.type == pygame.KEYDOWN:
                break



import abc
from collections.abc import Sequence

from sneks.engine.config.instantiation import config
from sneks.engine.core.action import Action
from sneks.engine.core.bearing import Bearing
from sneks.engine.core.cell import Cell
from sneks.engine.core.direction import Direction


class Snek(abc.ABC):
    """
    Base snek from which submission sneks derive. This snek has no behavior,
    but derivations should implement ``get_next_action()`` to provide it.
    """

    #: Head snek represented as Cell
    head: Cell = Cell(0, 0)
    #: Bearing of the snek
    bearing: Bearing = Bearing()
    #: Set of currently occupied cells on the game board that the snek can see
    occupied: frozenset[Cell] = frozenset()

    def get_next_action(self) -> Action:
        """
        Method that determines which action the snek should take next.

        :return: the next action for the snek to take
        """
        raise NotImplementedError

    def get_head(self) -> Cell:
        """
        Helper method to return the head of the snek.

        :return: the cell representing the head of the snek
        """
        return self.head

    def get_bearing(self) -> Bearing:
        """
        Helper method to return the bearing of the snek.

        :return: the bearing representing the directional speeds of the snek
        """
        return self.bearing

    def get_occupied(self) -> frozenset[Cell]:
        """
        Helper method to return all occupied cells that the snek can see.
        This includes both your snek's position, all other sneks'
        positions, and space occupied by asteroids.

        This can be used in your ``get_next_action()`` to check if a cell
        you are planning on moving to is already taken. Example::

            potential_next_cell = self.get_head().get_up()
            if potential_next_cell in self.get_occupied():
                # potential_next_cell is already taken
            else:
                # potential_next_cell is free

        :return: the set of occupied cells on the game board
        """
        return self.occupied

    def look(self, direction: Direction) -> int:
        """
        Look in a direction from the snek's head and get the distance to the closest obstacle.
        An obstacle could either be an occupied cell or the game board's border.

        >>> self.get_head()
        Cell(0, 0)
        >>> self.look(Direction.LEFT)
        0

        :param direction: the direction to look
        :return: the distance until the closest obstacle in the specified direction
        """

        current = self.get_head().get_neighbor(direction)
        current_distance = 1

        while (
            current not in self.occupied
            and current_distance <= config.game.vision_range
        ):
            current = current.get_neighbor(direction)
            current_distance += 1

        return current_distance - 1

    def get_direction_to_destination(
        self,
        destination: Cell,
        directions: Sequence[Direction] = (
            Direction.UP,
            Direction.DOWN,
            Direction.LEFT,
            Direction.RIGHT,
        ),
    ) -> Direction:
        """
        Get the next direction to travel in order to reach the destination
        from a set of specified directions (default: all directions).

        When multiple directions have the same resulting distance, the chosen
        direction is determined by the order provided, with directions coming
        first having precedence.

        For example, to get the direction the snek should travel to close the
        most distance between itself and a cell 5 columns and 9 rows away,
        this method could be used like::

            self.get_direction_to_destination(Cell(5, 9))

        :param destination: the cell to travel towards
        :param directions: the directions to evaluate in order
        :return: the direction to travel that will close the most distance
        """

        return min(
            directions,
            key=lambda direction: self.get_head()
            .get_neighbor(direction)
            .get_distance(destination),
        )



import math

from sneks.engine.config.instantiation import config
from sneks.engine.core.cell import Cell
from sneks.engine.core.direction import Direction


def get_absolute_neighbor(cell: Cell, direction: Direction) -> Cell:
    if direction is Direction.UP:
        return get_absolute_by_offset(cell=cell, y_offset=1)
    elif direction is Direction.DOWN:
        return get_absolute_by_offset(cell=cell, y_offset=-1)
    elif direction is Direction.LEFT:
        return get_absolute_by_offset(cell=cell, x_offset=-1)
    elif direction is Direction.RIGHT:
        return get_absolute_by_offset(cell=cell, x_offset=1)
    else:
        raise ValueError("direction not valid")


def get_absolute_by_offset(*, cell: Cell, x_offset: int = 0, y_offset: int = 0) -> Cell:
    return Cell(
        (cell.x + x_offset) % config.game.columns,
        (cell.y + y_offset) % config.game.rows,
    )


def get_relative_to(cell: Cell, other: Cell) -> Cell:
    """
    Returns the relative cell in relation to "other". Other is likely the head when this is called,
    since that's what the coordinates are referenced on for the snek implementation.
    """
    return Cell(
        int(math.fmod((cell.x - other.x), config.game.columns)),
        int(math.fmod((cell.y - other.y), config.game.rows)),
    )



from sneks.engine.config.instantiation import config
from sneks.engine.core.action import Action
from sneks.engine.core.bearing import Bearing


class Velocity:
    azimuth: Bearing

    def __init__(self):
        self.azimuth = Bearing()

    def add(self, action: Action) -> None:
        match action:
            case Action.UP:
                self.azimuth.up = min(
                    config.game.directional_speed_limit, self.azimuth.up + 1
                )
            case Action.DOWN:
                self.azimuth.up = max(
                    -1 * config.game.directional_speed_limit, self.azimuth.up - 1
                )
            case Action.LEFT:
                self.azimuth.right = max(
                    -1 * config.game.directional_speed_limit, self.azimuth.right - 1
                )
            case Action.RIGHT:
                self.azimuth.right = min(
                    config.game.directional_speed_limit, self.azimuth.right + 1
                )



from sneks.engine.config.instantiation import config
from sneks.engine.core.action import Action
from sneks.engine.core.cell import Cell
from sneks.engine.engine import registrar, runner
from sneks.engine.engine.mover import NormalizedScore
from sneks.engine.interface.snek import Snek


def test_basic_functionality() -> None:
    submissions = registrar.get_submissions()
    assert len(submissions) == 1
    snek: Snek = submissions[0].snek
    snek.occupied = frozenset((Cell(1, 1),))
    assert snek.get_next_action() in Action


def test_extended_functionality() -> None:
    assert config.graphics is not None
    config.graphics.display = False
    config.turn_limit = 100
    scores = runner.main()
    assert scores is not None
    assert len(scores) == 10
    for score in scores:
        assert isinstance(score, NormalizedScore)
        assert score.crashes == 0
        assert score.distance == 0
        assert score.raw.crashes >= 0
        assert score.raw.distance >= 0


def test_multiple_functionality() -> None:
    assert config.graphics is not None
    config.graphics.display = False
    config.turn_limit = 200
    config.registrar_submission_sneks = 10
    scores = runner.main()
    assert scores is not None
    assert len(scores) == 100
    for score in scores:
        assert isinstance(score, NormalizedScore)
        assert 0 <= score.crashes <= 1
        assert 0 <= score.distance <= 1
        assert score.raw.crashes >= 0
        assert score.raw.distance >= 0



from enum import Enum

from sneks.engine.core.action import Action


class Direction(Enum):
    """
    An enumeration to provide directional values
    """

    UP = "up"
    DOWN = "down"
    LEFT = "left"
    RIGHT = "right"

    def get_opposite(self) -> "Direction":
        """
        Gets the opposite direction of this direction

        :return: the opposite direction
        """

        if self == Direction.UP:
            return Direction.DOWN
        elif self == Direction.DOWN:
            return Direction.UP
        elif self == Direction.LEFT:
            return Direction.RIGHT
        else:
            return Direction.LEFT

    def get_action(self) -> Action:
        """
        Gets the corresponding action for the direction

        :return: the action that corresponds to the direction
        """

        return Action(self.value)



class Bearing:
    """
    Represents the directional speeds of a snek. The speeds are represented
    as integers to show the vertical and horizontal speeds in cells per
    game tick.

    Sneks start with a bearing of ``(0, 0)``. For example, if the action from
    ``get_next_action()`` is ``Action.UP``, the snek will increase the speed vertically
    by one, so on the next game step the bearing will be ``(0, 1)``, and the snek will
    have moved one cell.
    """

    up: int  #:
    right: int  #:

    def __init__(self, up: int = 0, right: int = 0) -> None:
        self.up = up
        self.right = right

    def __repr__(self) -> str:
        return f"({self.right}, {self.up})"



from enum import Enum


class Action(Enum):
    """
    An enumeration to provide action values. The action represents how the snek
    should accelerate. The bearing of the snek is the current speed, and an action
    will be applied to the current bearing.
    """

    MAINTAIN = "maintain"
    UP = "up"
    DOWN = "down"
    LEFT = "left"
    RIGHT = "right"



import abc
from collections import deque
from dataclasses import dataclass

from sneks.engine.core.bearing import Bearing
from sneks.engine.core.cell import Cell
from sneks.engine.core.direction import Direction
from sneks.engine.engine.cells import get_absolute_by_offset, get_absolute_neighbor
from sneks.engine.engine.velocity import Velocity
from sneks.engine.interface.snek import Snek


@dataclass(frozen=True)
class NormalizedScore:
    crashes: float
    distance: float
    raw: "Score"

    def total(self) -> float:
        return self.crashes + self.distance

    def __repr__(self):
        return (
            f"crashes': {self.crashes:.4f}, distance': {self.distance:.4f} "
            f"crashes: {self.raw.crashes:2d}, distance: {self.raw.distance:.2f} "
            f"name: {self.raw.name}"
        )


@dataclass(frozen=True)
class Score:
    name: str
    crashes: int
    distance: float

    def normalize(self, min_score: "Score", max_score: "Score") -> NormalizedScore:
        return NormalizedScore(
            crashes=(min_score.crashes - self.crashes)
            / max(1, (min_score.crashes - max_score.crashes)),
            distance=(self.distance - min_score.distance)
            / max(1.0, (max_score.distance - min_score.distance)),
            raw=self,
        )


class BaseMover(abc.ABC):
    def move(self) -> None:
        raise NotImplementedError


class Mover(BaseMover):
    name: str
    color: tuple[int, int, int]
    snek: Snek
    max_age: int = 0
    crashes: int = 0
    distance: float = 0.0
    tail: deque[Cell]
    head: Cell
    velocity: Velocity

    def __init__(self, name: str, head: Cell, snek: Snek, color: tuple[int, int, int]):
        self.name = name
        self.color = color
        self.snek = snek

        self.reset(head=head)

    def reset(self, head: Cell) -> None:
        self.head = head
        self.tail = deque()
        self.velocity = Velocity()
        self.tail.append(head)

    def get_head(self) -> Cell:
        return self.head

    def move(self):
        next_action = self.snek.get_next_action()
        self.velocity.add(next_action)

        next_head = self.get_head()

        ups = abs(self.velocity.azimuth.up)
        rights = abs(self.velocity.azimuth.right)

        while ups > 0 or rights > 0:
            if ups > 0:
                next_direction = (
                    Direction.UP if self.velocity.azimuth.up > 0 else Direction.DOWN
                )
                next_head = get_absolute_neighbor(next_head, next_direction)
                self.tail.append(next_head)
                ups -= 1
            if rights > 0:
                next_direction = (
                    Direction.RIGHT
                    if self.velocity.azimuth.right > 0
                    else Direction.LEFT
                )
                next_head = get_absolute_neighbor(next_head, next_direction)
                self.tail.append(next_head)
                rights -= 1

        while len(self.tail) > max(
            1, 3 * (abs(self.velocity.azimuth.up) + abs(self.velocity.azimuth.right))
        ):
            self.tail.popleft()
        self.distance += self.head.get_distance(next_head)
        self.head = next_head

    def get_score(self) -> Score:
        return Score(
            name=self.name,
            crashes=self.crashes,
            distance=self.distance,
        )


class Asteroid(BaseMover):
    cells: set[Cell]
    tail: set[Cell]
    azimuth: Bearing

    def __init__(self, *, cells: list[Cell], azimuth: Bearing) -> None:
        self.cells = set(cells)
        self.tail = set()
        self.azimuth = azimuth

    def move(self) -> None:
        self.cells = {
            get_absolute_by_offset(
                cell=cell, x_offset=self.azimuth.right, y_offset=self.azimuth.up
            )
            for cell in self.cells
        }

        self.tail = set()

        if abs(self.azimuth.up) == 2:
            self.tail.update(
                get_absolute_by_offset(
                    cell=cell, x_offset=0, y_offset=-1 if self.azimuth.up >= 0 else 1
                )
                for cell in self.cells
            )

        if abs(self.azimuth.right) == 2:
            self.tail.update(
                get_absolute_by_offset(
                    cell=cell, x_offset=-1 if self.azimuth.right >= 0 else 1, y_offset=0
                )
                for cell in self.cells
            )



import importlib
import importlib.util
import pathlib
import sys
from dataclasses import dataclass
from importlib.machinery import ModuleSpec
from types import ModuleType

from sneks.engine.config.instantiation import config
from sneks.engine.interface.snek import Snek


@dataclass(frozen=True)
class Submission:
    name: str
    snek: Snek


def get_submissions() -> list[Submission]:
    sneks: list[Submission] = []
    snek_classes = get_submission_classes()
    for name, snek in snek_classes.items():
        if config.registrar_submission_sneks > 1:
            for i in range(config.registrar_submission_sneks):
                sneks.append(Submission(f"{name}{i}", snek()))  # type: ignore
        else:
            sneks.append(Submission(name, snek()))  # type: ignore
    return sneks


def get_submission_classes() -> dict[str, Snek]:
    results = {}
    submissions = set(
        p.parent for p in pathlib.Path(config.registrar_prefix).glob("**/submission.py")
    )
    for submission in submissions:
        name, snek = get_custom_snek(submission)
        if name is not None and snek is not None:
            results[name] = snek
    return results


def get_custom_snek(prefix: pathlib.Path) -> tuple[str | None, Snek | None]:
    name, module = load_module(prefix)
    if module is None:
        return None, None
    return name, module.CustomSnek


def load_module(prefix: pathlib.Path) -> tuple[str | None, ModuleType | None]:
    name, spec, module = get_module(prefix)
    if name is not None and spec is not None and module is not None:
        assert spec.loader is not None
        spec.loader.exec_module(module)
        sys.modules[name] = module
    return name, module


def get_module(
    prefix: pathlib.Path,
) -> tuple[str | None, ModuleSpec | None, ModuleType | None]:
    submission = get_submission(prefix)
    if submission is None:
        return None, None, None
    name = get_submission_name(submission)
    spec = importlib.util.spec_from_file_location(name, submission)
    module = importlib.util.module_from_spec(spec) if spec is not None else None
    return name, spec, module


def get_submission(prefix: pathlib.Path) -> pathlib.Path | None:
    files = get_submission_files(prefix)
    if files:
        return files[0]
    return None


def get_submission_files(prefix: pathlib.Path) -> list[pathlib.Path]:
    suffix = "submission.py"
    return list(prefix.glob(f"**/{suffix}"))


def get_submission_name(prefix: pathlib.Path) -> str:
    return str(prefix.relative_to(config.registrar_prefix).parent)



import itertools
import random
from collections import Counter
from operator import methodcaller

from sneks.engine.config.instantiation import config
from sneks.engine.core.bearing import Bearing
from sneks.engine.core.cell import Cell
from sneks.engine.core.direction import Direction
from sneks.engine.engine import cells, registrar
from sneks.engine.engine.mover import Asteroid, Mover, NormalizedScore, Score


class State:
    def __init__(self):
        self.cells: set[Cell] = {
            Cell(x, y)
            for x, y in itertools.product(
                range(config.game.columns), range(config.game.rows)
            )
        }
        self.snakes: list[Mover] = []
        self.ended_snakes: list[Cell] = []
        self.asteroids: list[Asteroid] = []
        self.steps: int = 0
        self.occupied: set[Cell] = set()

    def reset(self):
        self.steps = 0
        self.snakes = []
        self.ended_snakes = []
        self.asteroids = []
        self.occupied = set()
        sneks = registrar.get_submissions()
        sneks.sort(key=lambda s: s.name)
        color_index = 0
        color_index_delta = max(len(config.graphics.colors.snake) // len(sneks), 1)
        for snek in sneks:
            self.snakes.append(
                Mover(
                    name=snek.name,
                    head=self.get_random_free_cell(),
                    snek=snek.snek,
                    color=config.graphics.colors.snake[color_index],
                )
            )
            color_index = (color_index + color_index_delta) % len(
                config.graphics.colors.snake
            )
        bearing_options = list(
            range(
                -config.game.asteroid_directional_speed_limit,
                config.game.asteroid_directional_speed_limit + 1,
            )
        )
        for _ in range(config.game.asteroid_count):
            up = random.choice(bearing_options)
            right = random.choice(bearing_options)
            if up == 0 and right == 0:
                # ensure all asteroids are moving
                increment = 1 if random.getrandbits(1) else -1
                if random.getrandbits(1):
                    up += increment
                else:
                    right += increment
            self.asteroids.append(
                Asteroid(
                    cells=self.get_random_free_cell(cluster_size=random.randint(3, 12)),
                    azimuth=Bearing(up=up, right=right),
                )
            )
        self.set_board()

    def report(self) -> list[NormalizedScore]:
        scores = [s.get_score() for s in self.snakes]
        min_crashes = min(s.crashes for s in scores)
        max_crashes = max(s.crashes for s in scores)
        min_distance = min(s.distance for s in scores)
        max_distance = max(s.distance for s in scores)

        min_score = Score(name="min", crashes=max_crashes, distance=min_distance)
        max_score = Score(
            name="max",
            crashes=min_crashes,
            distance=max_distance,
        )

        normalized = [
            s.normalize(min_score=min_score, max_score=max_score) for s in scores
        ]
        normalized.sort(key=methodcaller("total"), reverse=True)

        return normalized

    def get_random_free_cell(self, *, cluster_size: int = 1):
        # TODO: split out the logic for building asteroids (i.e. clusters)
        options = self.cells.difference(self.get_occupied_cells())
        if options:
            shuffled = list(options)
            if cluster_size == 1:
                return random.choice(shuffled)
            random.shuffle(shuffled)
            directions = list(Direction)
            for candidate in shuffled:
                potential = set()
                potential.add(candidate)
                while len(potential) < cluster_size:
                    for p in potential:
                        next_candidate = p.get_neighbor(random.choice(directions))
                        if (
                            next_candidate in options
                            and next_candidate not in potential
                        ):
                            potential.add(next_candidate)
                            break
                return list(potential)
            return random.choice(shuffled)
        else:
            return None

    def get_occupied_cells(self) -> frozenset[Cell]:
        occupied = frozenset().union(s.get_head() for s in self.snakes)
        return occupied.union(*(a.cells for a in self.asteroids))

    def set_board(self):
        self.occupied = self.get_occupied_cells()

        for current_snake in self.snakes:
            head = current_snake.get_head()

            # build a grid around the head based on the vision range
            grid = {
                Cell(x, y)
                for x, y in itertools.product(
                    range(
                        head.x - config.game.vision_range,
                        head.x + config.game.vision_range,
                    ),
                    range(
                        head.y - config.game.vision_range,
                        head.y + config.game.vision_range,
                    ),
                )
            }

            # set the snek's occupied to occupied cells within grid
            current_snake.snek.occupied = frozenset(
                cells.get_relative_to(cell, head)
                for cell in grid.intersection(self.occupied)
                if cell.get_distance(head) < config.game.vision_range
            )

            current_snake.snek.bearing.up = current_snake.velocity.azimuth.up
            current_snake.snek.bearing.right = current_snake.velocity.azimuth.right

    def should_continue(self, turn_limit):
        return self.steps < turn_limit and self.snakes

    def end_snake(self, snake: Mover):
        self.ended_snakes.append(snake.get_head())
        snake.crashes += 1
        snake.reset(self.get_random_free_cell())

    def step(self):
        # move the heads
        for snake in self.snakes:
            snake.move()

        for asteroid in self.asteroids:
            asteroid.move()

        asteroid_cells = frozenset().union(*(a.cells | a.tail for a in self.asteroids))
        occupations = Counter(s.get_head() for s in self.snakes)

        to_end = []
        # determine ended snakes
        for snake in self.snakes:
            if occupations[snake.get_head()] > 1:
                # collided with another snake
                to_end.append(snake)
            elif not asteroid_cells.isdisjoint(
                list(snake.tail)[
                    -(
                        abs(snake.velocity.azimuth.up)
                        + abs(snake.velocity.azimuth.right)
                    ) :
                ]
            ):
                # collided with asteroid
                to_end.append(snake)

        for snake in to_end:
            self.end_snake(snake)

        self.set_board()
        self.steps += 1
